@@ -11,26 +11,29 @@ use Exporter 'import';
 use List::Util qw( any first );
 use Scalar::Util qw( looks_like_number );
 
-our @EXPORT = qw( DEBUG DEBUG_API $X $cfg $focus $windows %screens %xcb_events %xcb_events_ignore @screens
-    add_event_cb add_event_ignore hexnum init_extension replace_event_cb screen_by_xy pointer
+our @EXPORT = qw(
+    carp croak S_DEBUG DEBUG_API $X $cfg $focus $windows %screens %xcb_events %xcb_events_ignore @screens %atoms
+    add_event_cb add_event_ignore hexnum init_extension replace_event_cb screen_by_xy pointer %marked_windows
     $visible_min_x $visible_min_y $visible_max_x $visible_max_y $prevent_focus_in $prevent_enter_notify $cpu_saver
-    focus_prev_push focus_prev_remove focus_prev_get prevent_focus_in prevent_enter_notify
+    focus_prev_push focus_prev_remove focus_prev_get prevent_focus_in prevent_enter_notify $cached_classes atom
     );
 
-# Set after parsing config
-sub DEBUG;
-sub DEBUG_API;
+# NOTE all the debug functions are defined in Config.pm
+push @EXPORT, "DEBUG$_" for 1..9;
 
 our $X;
+our %atoms;
 our $cfg;
 our $cpu_saver = 0.1; # number of seconds to sleep before events processing (100ms by default)
 our $focus;
 our $windows = {};
+our $cached_classes = {};
 our %screens;
 our %xcb_events;
 our %xcb_events_ignore;
 our @screens;
 our ($visible_min_x, $visible_min_y, $visible_max_x, $visible_max_y);
+our %marked_windows;
 
 # Sometimes we want to ignore FocusIn (see Mouse/ENTER_NOTIFY and Executor/tag_select)
 our $prevent_focus_in;
@@ -58,11 +61,11 @@ sub replace_event_cb($id, $sub) {
 
 sub init_extension($name, $first_event) {
     my $ext = $X->query_extension_reply($X->query_extension(length($name), $name)->{sequence});
-    die "$name extension not available" unless $ext->{present};
+    croak "$name extension not available" unless $ext->{present};
 
     # We can skip this part unless we're interested getting event
     return unless defined $first_event;
-    die "Could not get $name first_event" unless $$first_event = $ext->{first_event};
+    croak "Could not get $name first_event" unless $$first_event = $ext->{first_event};
 }
 
 # focus_prev helpers
@@ -83,11 +86,11 @@ sub focus_prev_get() {
 }
 
 # Preventor functions to avoid code copy-pasting
-sub prevent_focus_in($timeout = 0.2) {
+sub prevent_focus_in($timeout = 0.11) {
     $prevent_focus_in = AE::timer $timeout, 0, sub { $prevent_focus_in = undef };
 }
 
-sub prevent_enter_notify($timeout = 0.2) {
+sub prevent_enter_notify($timeout = 0.11) {
     $prevent_enter_notify = AE::timer $timeout, 0, sub { $prevent_enter_notify = undef };
 }
 
@@ -105,4 +108,10 @@ sub pointer($wid = $X->root->id) {
     $X->query_pointer_reply($X->query_pointer($wid)->{sequence}) // {};
 }
 
+# Caching function to resolve and create atoms
+sub atom($name) {
+    return $atoms{$name} if defined $atoms{$name};
+    $atoms{$name} = $X->intern_atom_reply($X->intern_atom(0, length($name), $name)->{sequence})->{atom};
+}
+
 1;
