@@ -6,7 +6,6 @@ use strict;
 use warnings;
 use feature 'signatures';
 
-use Carp;
 use X11::XCB ':all';
 use X11::korgwm::Common;
 require X11::korgwm::Config;
@@ -105,7 +104,8 @@ sub init {
             $cpu_saver = 0.0001;
             replace_event_cb(MOTION_NOTIFY, \&_motion_resize);
         } else {
-            croak "We got unexpected mouse event, detail:" . $evt->{detail};
+            # Ignore other buttons (2 = middle, 4 = scroll down, 5 = scroll up)
+            carp "We got unexpected mouse event, detail:" . $evt->{detail};
         }
     });
 
