@@ -8,7 +8,6 @@ use feature 'signatures';
 
 use List::Util qw( first );
 use Scalar::Util qw( weaken );
-use Carp;
 use X11::XCB ':all';
 use X11::korgwm::Common;
 use X11::korgwm::Layout;
@@ -63,9 +62,13 @@ sub hide($self) {
     $X->flush();
 }
 
-sub show($self) {
+# Makes the $self tag visible
+# Options:
+# - noselect => do not call select() to avoid warp_pointer()
+sub show($self, %opts) {
     # Redefine layout if needed
     $self->{layout} //= X11::korgwm::Layout->new();
+    DEBUG8 and carp "tag->show($self, @{[%opts]})";
 
     # Map all windows from the tag
     my ($w, $h, $x, $y) = @{ $self->{screen} }{qw( w h x y )};
@@ -101,14 +104,18 @@ sub show($self) {
     # Handle focus change
     $focus->{screen} = $self->{screen};
     my $focus_win = $self->{screen}->{focus};
-    if (defined $focus_win and exists $focus_win->{on_tags}->{$self} ) {
+    if (defined $focus_win and exists $focus_win->{on_tags}->{$self}) {
         # If this window is focused on this tag, just give it a focus
         $focus_win->focus();
     } else {
         # Try to focus previously focused window (or any window)
-        my $win = $self->{focus} || $self->first_window();
-        # XXX maybe drop focus otherwise? UPD: no. I want visible window to be focused this case
-        $win->focus() if $win;
+        if (my $win = $self->{focus} || $self->first_window()) {
+            if ($win->{floating} or $win->{maximized} or $opts{noselect}) {
+                $win->focus(); # floating, maximized, always_on
+            } else {
+                $win->select(); # tiled
+            }
+        }
     }
 
     $X->flush();
